/**
 *
 *  @file kernels.cu
 *
 *  @brief cuThomasBatch kernel implementaion.
 *
 *  cuThomasBatch is a software package provided by
 *  Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
 *
 *  @author Ivan Martinez-Perez ivan.martinez@bsc.es
 *  @author Pedro Valero-Lara   pedro.valero@bsc.es
 *
 **/
#include <stdlib.h>
#include <assert.h>
#include <cusolverSp.h>
#include <cuda_runtime_api.h>
#include <iostream>
#include <stdio.h>
#include "aux_func.h"
#include "utils.hpp"
#include <omp.h>
#include <cuda_runtime.h>
#include "cusparse.h"
#include "cuThomasBatch.h"

    double init;

    int chunk;
    int inter;
    int normal;

    double *u_seq;
    double *u_cuSparse_host;
    double *u_Thomas_host;
    double *u_device;
    double *u_input;

    double *d_seq;
    double *d_cuSparse_host;
    double *d_Thomas_host;
    double *d_device;
    double *d_input;

    double *l_seq;
    double *l_cuSparse_host;
    double *l_Thomas_host;
    double *l_device;
    double *l_input;

    double *rhs_seq;
    double *rhs_cuSparse_host;
    double *rhs_Thomas_host;
    double *rhs_device;
    double *rhs_input;

    double *rhs_seq_output;
    double *rhs_cuSparse_output;
    double *rhs_Thomas_output;
    double *rhs_seq_interleave;

// Kernels 
void solve_seq(double* l, double* d, double* u, double* rhs, int n) {

    int first,last;
	//printf("BATCHCOUNT = %d, SIZE = %d\n",N,n);
    for (int j = 0; j < N; ++j)
    {
        first = j*n;
        last = first + n;
        last--; 

        u[first] /= d[first];
        rhs[first] /= d[first];

        for (int i = first+1; i < last; i++) {
            u[i] /= d[i] - l[i]*u[i-1];
            rhs[i] = (rhs[i] - l[i]*rhs[i-1]) / (d[i] - l[i]*u[i-1]);
        }

        rhs[last] = (rhs[last] - l[last]*rhs[last-1]) / (d[last] - l[last]*u[last-1]);

        for (int i = last-1; i >= first; i--) {
            rhs[i] -= u[i]*rhs[i+1];
        }
    }

}

void solve_multicore(double* l, double* d, double* u, double* rhs, int n) {

    int first,last;
    #pragma omp parallel for shared(l,d,u,rhs)
    for (int j = 0; j < N; ++j)
    {
        first = j*n;
        last = first + n;

        last--; 

        u[first] /= d[first];
        rhs[first] /= d[first];

        for (int i = first+1; i < last; i++) {
            u[i] /= d[i] - l[i]*u[i-1];
            rhs[i] = (rhs[i] - l[i]*rhs[i-1]) / (d[i] - l[i]*u[i-1]);
        }

        rhs[last] = (rhs[last] - l[last]*rhs[last-1]) / (d[last] - l[last]*u[last-1]);
    
        for (int i = last-1; i >= first; i--) {
            rhs[i] -= u[i]*rhs[i+1];
        }
    }

}

// Setup functions 
void allocateGlobalVariables(ThomasMatrix const& params){  

    u_seq = (double*) malloc(N*params.M*sizeof(double));
    u_cuSparse_host = (double*) malloc(N*params.M*sizeof(double));
    u_Thomas_host =  (double*) malloc(N*params.M*sizeof(double));
    u_input = (double*) malloc(N*params.M*sizeof(double));

    d_seq = (double*) malloc(N*params.M*sizeof(double));
    d_cuSparse_host = (double*) malloc(N*params.M*sizeof(double));
    d_Thomas_host =  (double*) malloc(N*params.M*sizeof(double));
    d_input = (double*) malloc(N*params.M*sizeof(double));

    l_seq = (double*) malloc(N*params.M*sizeof(double));
    l_cuSparse_host = (double*) malloc(N*params.M*sizeof(double));
    l_Thomas_host =  (double*) malloc(N*params.M*sizeof(double));
    l_input = (double*) malloc(N*params.M*sizeof(double));

    rhs_seq = (double*) malloc(N*params.M*sizeof(double));
    rhs_cuSparse_host = (double*) malloc(N*params.M*sizeof(double));
    rhs_Thomas_host = (double*) malloc(N*params.M*sizeof(double));
    rhs_input = (double*) malloc(N*params.M*sizeof(double));


    rhs_seq_output = (double*) malloc(N*params.M*sizeof(double));
    rhs_cuSparse_output=(double*) malloc(N*params.M*sizeof(double));
    rhs_Thomas_output=(double*) malloc(N*params.M*sizeof(double));
    rhs_seq_interleave = (double*) malloc(N*params.M*sizeof(double));

}

void allocateGPUVariables(ThomasMatrix const& params){

    cudaMalloc(&u_device ,N*params.M*sizeof(double));

    cudaMalloc(&d_device ,N*params.M*sizeof(double));

    cudaMalloc(&l_device ,N*params.M*sizeof(double));

    cudaMalloc(&rhs_device ,N*params.M*sizeof(double));

}

void freeGlobalVariables(){

    free(u_seq);  
    free(u_cuSparse_host);
    free(u_Thomas_host);
    
    free(d_seq);  
    free(d_cuSparse_host);
    free(d_Thomas_host);

    free(l_seq);  
    free(l_cuSparse_host);
    free(l_Thomas_host);

    free(rhs_seq);  
    free(rhs_cuSparse_host);
    free(rhs_Thomas_host);

}

void initSimpleGlobalVariables(ThomasMatrix const& params){
    
    for (int i = 0; i < N; ++i)
    {
       for (int j = 0; j < params.M; ++j)
        {
            u_seq[(i * params.M) + j] = params.a[j];
            u_cuSparse_host[(i * params.M) + j] = params.a[j];
            u_input[(i * params.M) + j] = params.a[j];

            d_seq[(i * params.M) + j] = params.d[j];
            d_cuSparse_host[(i * params.M) + j] = params.d[j];
            d_input[(i * params.M) + j] = params.d[j];

            l_seq[(i * params.M) + j] = params.b[j];
            l_cuSparse_host[(i * params.M) + j] = params.b[j];
            l_input[(i * params.M) + j] = params.b[j];

            rhs_seq[(i * params.M) + j] = params.rhs[j];
            rhs_cuSparse_host[(i * params.M) + j] = params.rhs[j];
            rhs_input[(i * params.M) + j] = params.rhs[j];

        }
    }

}

void initInterleavedGlobalVariables(ThomasMatrix const& params){

    for (int i = 0; i < params.M; ++i)
    {
        for (int j = 0; j < N; ++j)
        {
            u_Thomas_host[i*N+j] = u_input[j*params.M+i];
            l_Thomas_host[i*N+j] = l_input[j*params.M+i];
            d_Thomas_host[i*N+j] = d_input[j*params.M+i];
            rhs_Thomas_host[i*N+j] = rhs_input[j*params.M+i];

            rhs_seq_interleave[i*N+j] = rhs_seq_output[j*params.M+i];
                
        }
    }

}

// Wrapers 
void execSeq(ThomasMatrix const& params){

   init = time_wtime();
    solve_seq(
        l_seq,
        d_seq,
        u_seq,
        rhs_seq,
        params.M
    );
    printf("        CPU SEQ Time(s) %e\n", time_wtime()-init);

    for (int i = 0; i < params.M*N; ++i)
        rhs_seq_output[i] = rhs_seq[i];

}

void execMulti(ThomasMatrix const& params){

   init = time_wtime();
    solve_multicore(
        l_seq,
        d_seq,
        u_seq,
        rhs_seq,
        params.M
    );
    printf("        CPU MULT Time(s) %e ", time_wtime()-init);
    calcError(rhs_seq_output,rhs_seq,N*params.M);

}

void execCuThomas(ThomasMatrix const& params){
    
    cudaMemcpy(u_device,u_Thomas_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);

    cudaMemcpy(l_device,l_Thomas_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);

    cudaMemcpy(d_device,d_Thomas_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);

    cudaMemcpy(rhs_device,rhs_Thomas_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);
    
    init = time_wtime();
    cuThomasBatch<<<(N/BlockSize)+1, BlockSize>>>
    (
        l_device, d_device, u_device, rhs_device,
        params.M,
        N
    );
    
    cudaDeviceSynchronize();
    printf("        cuThomasBatch Time(s) %e   ", time_wtime()-init);
    cudaMemcpy(rhs_Thomas_output,rhs_device,N*params.M*sizeof(double),cudaMemcpyDeviceToHost);
    calcError(rhs_seq_interleave,rhs_Thomas_output,N*params.M);

}

void execCuSparse(ThomasMatrix const& params){

    cudaMemcpy(u_device,u_cuSparse_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);
    cudaMemcpy(l_device,l_cuSparse_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);
    cudaMemcpy(d_device,d_cuSparse_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);
    cudaMemcpy(rhs_device,rhs_cuSparse_host,N*params.M*sizeof(double),cudaMemcpyHostToDevice);

    cusparseStatus_t status;
    cusparseHandle_t myhandle=0;

    cusparseCreate(&myhandle);
	init = time_wtime();
    status = cusparseDgtsvStridedBatch(myhandle,params.M,l_device,d_device,u_device,rhs_device,N,params.M);

    printf("        gtsvStridedBatch Time(s) %e", time_wtime()-init);


    if (status != CUSPARSE_STATUS_SUCCESS) {
        printf("Tridiagonal solver failed\n");
    }
    
    cudaMemcpy(rhs_cuSparse_output,rhs_device,N*params.M*sizeof(double),cudaMemcpyDeviceToHost);
    calcError(rhs_seq_output,rhs_cuSparse_output,N*params.M);

}

void execCuSolverQR(int SIZE, int BATCHCOUNT, int *CSRROW, 
					int *CSRCOLUMN, double *CSRVALUE, double *RHS,
					ThomasMatrix const& PARAMS){
 	cusolverSpHandle_t cusolverH = NULL;
// GPU does batch QR
    csrqrInfo_t info = NULL;
    cusparseMatDescr_t descrA = NULL;

    cusparseStatus_t cusparse_status = CUSPARSE_STATUS_SUCCESS;
    cusolverStatus_t cusolver_status = CUSOLVER_STATUS_SUCCESS;
    cudaError_t cudaStat1 = cudaSuccess;
    cudaError_t cudaStat2 = cudaSuccess;
    cudaError_t cudaStat3 = cudaSuccess;
    cudaError_t cudaStat4 = cudaSuccess;
    cudaError_t cudaStat5 = cudaSuccess;

// GPU does batch QR
// d_A is CSR format, d_csrValA is of size nnzA*batchSize
// d_x is a matrix of size batchSize * m
// d_b is a matrix of size batchSize * m
    int *d_csrRowPtrA = NULL;
    int *d_csrColIndA = NULL;
    double *d_csrValA = NULL;
    double *d_b = NULL; // batchSize * m
    double *d_x = NULL; // batchSize * m

    size_t size_qr = 0;
    size_t size_internal = 0;
    void *buffer_qr = NULL; // working space for numerical factorization

/*      | 1                |
 *  A = |       2          |
 *      |            3     |
 *      | 0.1  0.1  0.1  4 |
 *  CSR of A is based-1
 *
 *  b = [1 1 1 1]
 */ 

    /*const int m = 4 ;
    const int nnzA = 7;
    const int csrRowPtrA[m+1]  = { 1, 2, 3, 4, 8};
    const int csrColIndA[nnzA] = { 1, 2, 3, 1, 2, 3, 4};
    const double csrValA[nnzA] = { 1.0, 2.0, 3.0, 0.1, 0.1, 0.1, 4.0};
    const double b[m] = {1.0, 1.0, 1.0, 1.0};
    const int batchSize = 17;*/

    const int m = SIZE ;
    const int nnzA = 3*(m-2)+4;
	int *csrRowPtrA = (int*) malloc((m+1)*sizeof(int));
    int *csrColIndA = (int*) malloc(nnzA*sizeof(int));
    double *csrValA = (double*) malloc(nnzA*sizeof(double)); 
    double *b = (double*) malloc(m*sizeof(double));
    const int batchSize = BATCHCOUNT;

	memcpy(csrRowPtrA,CSRROW,(m+1)*sizeof(int));
	memcpy(csrColIndA,CSRCOLUMN,nnzA*sizeof(int));
	memcpy(csrValA,CSRVALUE,nnzA*sizeof(double));
	memcpy(b,RHS,m*sizeof(double));
	

    double *csrValABatch = (double*)malloc(sizeof(double)*nnzA*batchSize);
    double *bBatch       = (double*)malloc(sizeof(double)*m*batchSize);
    double *xBatch       = (double*)malloc(sizeof(double)*m*batchSize);
    assert( NULL != csrValABatch );
    assert( NULL != bBatch );
    assert( NULL != xBatch );

// step 1: prepare Aj and bj on host
//  Aj is a small perturbation of A
//  bj is a small perturbation of b
//  csrValABatch = [A0, A1, A2, ...]
//  bBatch = [b0, b1, b2, ...]
    for(int colidx = 0 ; colidx < nnzA ; colidx++){
        double Areg = csrValA[colidx];
        for (int batchId = 0 ; batchId < batchSize ; batchId++){
            double eps = ((double)((rand() % 100) + 1)) * 1.e-4;
            csrValABatch[batchId*nnzA + colidx] = Areg + eps;
        }  
    }

    for(int j = 0 ; j < m ; j++){
        double breg = b[j];
        for (int batchId = 0 ; batchId < batchSize ; batchId++){
            double eps = ((double)((rand() % 100) + 1)) * 1.e-4;
            bBatch[batchId*m + j] = breg + eps;
        }  
    }

// step 2: create cusolver handle, qr info and matrix descriptor
    cusolver_status = cusolverSpCreate(&cusolverH);
    assert (cusolver_status == CUSOLVER_STATUS_SUCCESS);

    cusparse_status = cusparseCreateMatDescr(&descrA); 
    assert(cusparse_status == CUSPARSE_STATUS_SUCCESS);

    cusparseSetMatType(descrA, CUSPARSE_MATRIX_TYPE_GENERAL);
    cusparseSetMatIndexBase(descrA, CUSPARSE_INDEX_BASE_ONE); // base-1

    cusolver_status = cusolverSpCreateCsrqrInfo(&info);
    assert(cusolver_status == CUSOLVER_STATUS_SUCCESS);

// step 3: copy Aj and bj to device
    cudaStat1 = cudaMalloc ((void**)&d_csrValA   , sizeof(double) * nnzA * batchSize);
    cudaStat2 = cudaMalloc ((void**)&d_csrColIndA, sizeof(int) * nnzA);
    cudaStat3 = cudaMalloc ((void**)&d_csrRowPtrA, sizeof(int) * (m+1));
    cudaStat4 = cudaMalloc ((void**)&d_b         , sizeof(double) * m * batchSize);
    cudaStat5 = cudaMalloc ((void**)&d_x         , sizeof(double) * m * batchSize);
    assert(cudaStat1 == cudaSuccess);
    assert(cudaStat2 == cudaSuccess);
    assert(cudaStat3 == cudaSuccess);
    assert(cudaStat4 == cudaSuccess);
    assert(cudaStat5 == cudaSuccess);

    cudaStat1 = cudaMemcpy(d_csrValA   , csrValABatch, sizeof(double) * nnzA * batchSize, cudaMemcpyHostToDevice);
    cudaStat2 = cudaMemcpy(d_csrColIndA, csrColIndA, sizeof(int) * nnzA, cudaMemcpyHostToDevice);
    cudaStat3 = cudaMemcpy(d_csrRowPtrA, csrRowPtrA, sizeof(int) * (m+1), cudaMemcpyHostToDevice);
    cudaStat4 = cudaMemcpy(d_b, bBatch, sizeof(double) * m * batchSize, cudaMemcpyHostToDevice);
    assert(cudaStat1 == cudaSuccess);
    assert(cudaStat2 == cudaSuccess);
    assert(cudaStat3 == cudaSuccess);
    assert(cudaStat4 == cudaSuccess);

// step 4: symbolic analysis
    cusolver_status = cusolverSpXcsrqrAnalysisBatched(
        cusolverH, m, m, nnzA,
        descrA, d_csrRowPtrA, d_csrColIndA,
        info);
    assert(cusolver_status == CUSOLVER_STATUS_SUCCESS);

// step 5: prepare working space
    cusolver_status = cusolverSpDcsrqrBufferInfoBatched(
         cusolverH, m, m, nnzA,
         descrA, d_csrValA, d_csrRowPtrA, d_csrColIndA,
         batchSize,
         info,
         &size_internal,
         &size_qr);
    assert(cusolver_status == CUSOLVER_STATUS_SUCCESS);

    //printf("numerical factorization needs internal data %lld bytes\n", (long long)size_internal);      
    //printf("numerical factorization needs working space %lld bytes\n", (long long)size_qr);      

    cudaStat1 = cudaMalloc((void**)&buffer_qr, size_qr);
    assert(cudaStat1 == cudaSuccess);

// step 6: numerical factorization
// assume device memory is big enough to compute all matrices.
	
    init = time_wtime();
    cusolver_status = cusolverSpDcsrqrsvBatched(
        cusolverH, m, m, nnzA,
        descrA, d_csrValA, d_csrRowPtrA, d_csrColIndA,
        d_b, d_x,
        batchSize,
        info,
        buffer_qr);
	cudaDeviceSynchronize();
    printf("        SpDcsrqrsvBatched Time(s) %e   ", time_wtime()-init);
    cudaStat1 = cudaMemcpy(xBatch, d_x, sizeof(double)*m*batchSize, cudaMemcpyDeviceToHost);
	calcError(rhs_seq_output,xBatch,m*batchSize);

    assert(cusolver_status == CUSOLVER_STATUS_SUCCESS);

// step 7: check residual 
// xBatch = [x0, x1, x2, ...]
    assert(cudaStat1 == cudaSuccess);

    const int baseA = (CUSPARSE_INDEX_BASE_ONE == cusparseGetMatIndexBase(descrA))? 1:0 ;

    for(int batchId = 0 ; batchId < batchSize; batchId++){
        // measure |bj - Aj*xj|
        double *csrValAj = csrValABatch + batchId * nnzA;
        double *xj = xBatch + batchId * m;
        double *bj = bBatch + batchId * m;
        // sup| bj - Aj*xj|
        double sup_res = 0;
        for(int row = 0 ; row < m ; row++){
            const int start = csrRowPtrA[row ] - baseA;
            const int end   = csrRowPtrA[row+1] - baseA;
            double Ax = 0.0; // Aj(row,:)*xj
            for(int colidx = start ; colidx < end ; colidx++){
                const int col = csrColIndA[colidx] - baseA;
                const double Areg = csrValAj[colidx];
                const double xreg = xj[col];
                Ax = Ax + Areg * xreg;
            }
            double r = bj[row] - Ax;
            sup_res = (sup_res > fabs(r))? sup_res : fabs(r);
        }
        //printf("batchId %d: sup|bj - Aj*xj| = %E \n", batchId, sup_res);
    }

    for(int batchId = 0 ; batchId < batchSize; batchId++){
        //double *xj = xBatch + batchId * m;
        for(int row = 0 ; row < m ; row++){
            //printf("x%d[%d] = %E\n", batchId, row, xj[row]);
        } 
        //printf("\n");
    }

    return;
} 
